/* cls - list contents of directories in color
   Copyright (C) 1998 Rex Barzee

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "cls.h"
#include "error.h"
#include "format.h"

#include <dirent.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

const char ident[] =
  "@(#)color `ls' for dtterm and hpterm, version 1.8 - barzee@fc.hp.com";

termtype f_term;
bool f_listdot;     /* list files beginning with . */
bool f_listcurrent; /* list current and parent directories, i.e. . and .. */
bool f_listdir;     /* list actual directory, not contents */
bool f_recursive;

/* print nongraphic characters as ascii, octal, or question marks */
NonGraphic f_nongraphic = Ascii;

PrintForm f_printform = ColumnDown;
bool f_escape;   /* output escape sequences */
bool f_inode;    /* print inode number */
bool f_blocks;   /* print block size of file */
bool f_total;    /* print total number of blocks */
bool f_noowner;  /* long form without user */
bool f_nogroup;  /* long listing without group */
bool f_numbers;  /* long listing with user and group numbers */
bool f_type;     /* print type character for non-regular files */
bool f_link;     /* List the file or directory to which the link
                    refers rather than the link itself. */

bool f_reversesort;  /* reverse whatever sort is used */
bool f_timesort;     /* if true, sort by time first, else sort by name */
UseTime f_usetime = Modification;

static int columns = 80;
static const char options[] = "1ACEFLRabcdgilmnoqrstux";
static char path_buf[MAX_PATH + 1];
static char link_buf[MAX_PATH + 1];
static bool output;  /* If true, we must write a newline before writing
                        the next output. */


static void Usage() {
  fprintf(stderr, "Usage: %s [-%s] [files]\n", program, options);
  switch (Terminal(true)) {
    case Open:  FORMATS = OPEN_FORMATS;
                break;
    case Dtterm:  FORMATS = DT_FORMATS;
                  break;
    case Hpterm:  FORMATS = HP_FORMATS;
                  break;
  }
  fputs(
    "  cls version 1.8, Copyright (C) 1998 Rex Barzee\n"
    "  cls comes with ABSOLUTELY NO WARRANTY; for details see the GNU General\n"
    "  Public License version 2.  This is free software, and you are welcome\n"
    "  to redistribute it under certain conditions; for details see the GNU\n"
    "  General Public License version 2.\n"
    "\n"
    "  List contents of directories in color (see man ls).", stderr);
                                          fprintf(stderr, "  Directories are\n"
    "  %s; executables are %s; symbolic links are %s; fifos are %s;\n"
    "  and sockets are %s.",
    FORMATS[Directory + 2], FORMATS[Executable + 2], FORMATS[SymLink + 2],
    FORMATS[Fifo + 2], FORMATS[Socket + 2]);
                          fputs("  To determine terminal type (and thus\n"
    "  escape codes for colors), cls reads the \"TERM\" environment variable.\n"
    "  cls works in the following terminals: ansi_xterm, color_xterm, dtterm,\n"
    "  hpterm, kterm, nxterm, and rxvt.  Since cls doesn't print escape codes\n"
    "  if stdout is not a tty, in theory, you should be able to alias ls to\n"
    "  cls and have no problems.  However, please note that cls does NOT\n"
    "  support the following ls flags:  efp.  It does support:\n"
    "  1ACFLRabcdgilmnoqrstux.  If you want cls to produce escape codes\n"
    "  even though stdout isn't a tty, use -E, e.g. \"cls -E | more -v\"\n",
    stderr);
  exit(1);
}


/* If dir is not NULL, copy <dir>/<base> to path, else copy base to path.
 * dir must never be "". */
static char *pathname(char *path, const char *dir, const char *base) {
  char *p;

  if (dir) {
    strcpy(path, dir);
    p = strchr(path, '\0') - 1;
    if (*p++ != '/')
      *p++ = '/';
    strcpy(p, base);
  }
  else
    strcpy(path, base);
  return path;
}


/* Read and copy file system info about the file <path> into entry. */
static int ReadEntry(const char *path, const char *name, struct clsent *entry) {
  struct stat buf;
  int n;

  /* lstat() and stat() can take a long time, especially over NFS.  So
   * call them only if we have to.  We must always call lstat() so we
   * know what color to paint the filename. */
  if (lstat(path, &buf) != 0) {
    Warning("%s not found", name);
    return -1;
  }
  entry->len = NameLen(name);
  entry->name = strdup(name);
  entry->user  = GetUser(entry->uid = buf.st_uid);
  entry->group = GetGroup(entry->gid = buf.st_gid);
  entry->ino     = buf.st_ino;
  entry->mode    = buf.st_mode;
  entry->nlink   = buf.st_nlink;
  entry->rdev    = buf.st_rdev;
  entry->size    = buf.st_size;
  entry->atime   = buf.st_atime;
  entry->mtime   = buf.st_mtime;
  entry->ctime   = buf.st_ctime;
  entry->blksize = buf.st_blksize;
  entry->blocks  = buf.st_blocks;
  if ((entry->type = FileType(buf.st_mode)) != SymLink) {
    /* "name" is not a symbolic link. */
    entry->target_type = NoType;
    entry->target_name = NULL;
  }
  else if (f_printform == Long || f_link) {
    /* "name" is a symbolic link, and
     * we are listing the output in long format, or
     * we are listing the file or directory to which the link refers
     * rather than the link itself so we must call stat on the link
     * target name.  So we must stat the target of the link. */
    if ((n = readlink(path, link_buf, sizeof(link_buf))) > 0) {
      link_buf[n] = '\0';
      entry->target_name = strdup(link_buf);
    }
    if ((n = stat(path, &buf)) == 0) {
      entry->target_type = FileType(buf.st_mode);
      if (f_link) {
        entry->user  = GetUser(entry->uid = buf.st_uid);
        entry->group = GetGroup(entry->gid = buf.st_gid);
        entry->ino     = buf.st_ino;
        entry->mode    = buf.st_mode;
        entry->nlink   = buf.st_nlink;
        entry->rdev    = buf.st_rdev;
        entry->size    = buf.st_size;
        entry->atime   = buf.st_atime;
        entry->mtime   = buf.st_mtime;
        entry->ctime   = buf.st_ctime;
        entry->blksize = buf.st_blksize;
        entry->blocks  = buf.st_blocks;
      }
    }
    else if (f_link) {
      /* "name" is a symbolic link that points to a non-existent file,
       * "link_buf", so issue a warning. */
      Warning("%s", link_buf);
      return -1;
    }
  }
  return entry->len;
}


static int CopyEntries(char *argv[], int argc, struct
                       clsent *entries, int *num) {
  int i, j, len, max_len;

  for (i = j = max_len = 0;  j < argc;  j++) {
    if ((len = ReadEntry(argv[j], argv[j], &entries[i])) > 0) {
      if (len > max_len)
        max_len = len;
      i++;
    }
  }
  *num = i;
  return max_len;
}


static int ReadEntries(const char *name, DIR *dir,
                       struct clsent *entries, int *num) {
  struct dirent *dirent;
  int i, len, max_len;
  char *p;

  strcpy(path_buf, name);
  p = strchr(path_buf, '\0') - 1;
  if (*p++ != '/')
    *p++ = '/';
  i = max_len = 0;
  while ((dirent = readdir(dir)) != NULL) {
    if (!f_listdot && dirent->d_name[0] == '.')
      continue;
    if (!f_listcurrent &&
        (!strcmp(dirent->d_name, ".") || !strcmp(dirent->d_name, "..")))
      continue;
    strcpy(p, dirent->d_name);
    if ((len = ReadEntry(path_buf, dirent->d_name, &entries[i])) > 0) {
      if (len > max_len)
        max_len = len;
      i++;
    }
  }
  *num = i;
  return max_len;
}


static void FreeEntries(struct clsent *entries, int num) {
  while (num) {
    if (entries->name)
      free((void *)entries->name);
    if (entries->target_name)
      free((void *)entries->target_name);
    ++entries;
    --num;
  }
}


static void Traverse(const char *name) {
  DIR *dir;
  struct dirent *dirent;
  struct clsent *entry, *entries;
  int num, max_len;
  char cwd[MAX_PATH + 1];

  if ((dir = opendir(name)) == NULL) {
    Warning("can't read %s", name);
    return;
  }

  num = 0;
  while ((dirent = readdir(dir)) != NULL)
    ++num;

  entries = (struct clsent *)calloc(num, sizeof(*entries));
  rewinddir(dir);
  max_len = ReadEntries(name, dir, entries, &num);
  closedir(dir);

  if (!f_recursive) {
    qsort(entries, num, sizeof(*entries), SelectNonRecursSort());
    if (output) puts("");
    PrintFunc(entries, num, max_len, columns);
    output = true;
  }
  else {
    int i, num_nondirs;

    qsort(entries, num, sizeof(*entries), SelectRecursSort());

    /* Print all non directory entries. */
    for (num_nondirs = 0;  num_nondirs < num;  num_nondirs++)
      if (IsDirectory(&entries[num_nondirs]))
        break;
    if (num_nondirs) {
      if (output) puts("");
      PrintFunc(entries, num_nondirs, max_len, columns);
      output = true;
    }

    /* Recursively print contents of directory entries. */
    for (i = num_nondirs;  i < num;  i++) {
      entry = &entries[i];
      if (strcmp(entry->name, ".") != 0 && strcmp(entry->name, "..") != 0) {
        pathname(cwd, name, entries[i].name);
        if (output) puts("");
        PrintDir(cwd);
        output = false;
        Traverse(cwd);
      }
    }
  }
  FreeEntries(entries, num);
  free(entries);
}


static void ListCommandLine(int argc, char *argv[]) {
  struct clsent *entries;
  int i, num, max_len, num_nondirs;

  entries = (struct clsent *)calloc(argc, sizeof(*entries));
  max_len = CopyEntries(argv, argc, entries, &num);
  if (f_listdir) {
    qsort(entries, num, sizeof(*entries), SelectNonRecursSort());
    PrintFunc(entries, num, max_len, columns);
  }
  else {
    qsort(entries, num, sizeof(*entries), SelectRecursSort());

    /* Print all non directory entries. */
    for (num_nondirs = 0;  num_nondirs < num;  num_nondirs++)
      if (IsDirectory(&entries[num_nondirs]))
        break;
    if (f_printform == Long || num - num_nondirs > 0) {
      /* Output format is long or we will be traversing down directories,
       * so enable printing of total number of blocks. */
      f_total = true;
    }
    if (num_nondirs) {
      PrintFunc(entries, num_nondirs, max_len, columns);
      output = true;
    }

    /* Print all directory entries. */
    for (i = num_nondirs;  i < num;  i++) {
      if (output) puts("");

      /* Print the directory name only if more than one directory
         was specified on the command line. */
      if (num > 1)
        PrintDir(entries[i].name);

      output = false;
      Traverse(entries[i].name);
    }
  }
  FreeEntries(entries, argc);
  free(entries);
}


int main(int argc, char *argv[]) {
  static char *dotav[] = {".", NULL};
  int c;
  const char *p;

  if ((p = strrchr(argv[0], '/')) == NULL)
    program = argv[0];
  else
    program = p + 1;
  while ((c = getopt(argc, argv, options)) != EOF) {
    switch (c) {
      case '1':  f_printform = SingleColumn;
                 break;
      case 'A':  f_listdot = true;
                 break;
      case 'C':  f_printform = ColumnDown;
                 break;
      case 'E':  f_escape = true;
                 break;
      case 'F':  f_type = true;
                 break;
      case 'L':  f_link = true;
                 break;
      case 'R':  f_recursive = true;
                 f_listdir = false;
                 break;
      case 'a':  f_listdot = f_listcurrent = true;
                 break;
      case 'b':  f_nongraphic = Octal;
                 break;
      case 'c':  f_usetime = Status;
                 break;
      case 'd':  f_listdir = true;
                 f_recursive = false;
                 break;
      case 'g':  f_noowner = true;
                 f_printform = Long;
                 break;
      case 'i':  f_inode = true;
                 break;
      case 'l':  f_printform = Long;
                 break;
      case 'm':  f_printform = Stream;
                 break;
      case 'n':  f_numbers = true;
                 f_printform = Long;
                 break;
      case 'o':  f_nogroup = true;
                 f_printform = Long;
                 break;
      case 'q':  f_nongraphic = Question;
                 break;
      case 'r':  f_reversesort = true;
                 break;
      case 's':  f_blocks = true;
                 break;
      case 't':  f_timesort = true;
                 break;
      case 'u':  f_usetime = Access;
                 break;
      case 'x':  f_printform = ColumnAcross;
                 break;
      case '?':
      default:   Usage();
                 break;
    }
  }

  /* Root is -A automatically. */
  if (!getuid())
    f_listdot = true;

  f_term = Terminal(f_escape);
  if ((p = getenv("COLUMNS")) != NULL)
    columns = atoi(p);

  PrintInit();
  if (argc - optind > 0)
    ListCommandLine(argc - optind, argv + optind);
  else
    ListCommandLine(1, dotav);
  return 0;
}
