/* Print information (owner, group, inode, etc.) about files in various formats.
   Copyright (C) 1998 Rex Barzee

   This file is part of cls which lists contents of directories in color.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "cls.h"
#include "format.h"
#include <stdio.h>
#include <string.h>
#include <sys/stat.h>
#include <time.h>

PrintFcn PrintFunc;
NameFcn  PrintName;
NameFcn  NameLen;

static const char **TYPE_FORMATS, **LONG_FORMATS;

static const char *DT_TYPE_FORMATS[] = {
  NONE, NONE, NULL,
  NONE, NONE, NULL,
  DT_YELLOW,  DT_OFF_DIR,  NULL,
  DT_GREEN,   DT_OFF_EXEC, NULL,
  DT_RED,     DT_OFF_FIFO, NULL,
  DT_MAGENTA, DT_OFF_SOCK, NULL,
  DT_CYAN,    DT_OFF_LINK, NULL,
  DT_CYAN, DT_OFF,      NULL,
  DT_CYAN, DT_OFF_DIR,  NULL,
  DT_CYAN, DT_OFF_EXEC, NULL,
  DT_CYAN, DT_OFF_FIFO, NULL,
  DT_CYAN, DT_OFF_SOCK, NULL,
  DT_CYAN, DT_OFF_LINK, NULL,
};

static const char *HP_TYPE_FORMATS[] = {
  NONE, NONE, NULL,
  NONE, NONE, NULL,
  HP_YELLOW,  HP_OFF_DIR,  NULL,
  HP_GREEN,   HP_OFF_EXEC, NULL,
  HP_RED,     HP_OFF_FIFO, NULL,
  HP_MAGENTA, HP_OFF_SOCK, NULL,
  HP_CYAN,    HP_OFF_LINK, NULL,
  HP_CYAN, HP_OFF,      NULL,
  HP_CYAN, HP_OFF_DIR,  NULL,
  HP_CYAN, HP_OFF_EXEC, NULL,
  HP_CYAN, HP_OFF_FIFO, NULL,
  HP_CYAN, HP_OFF_SOCK, NULL,
  HP_CYAN, HP_OFF_LINK, NULL,
};

static const char *OPEN_TYPE_FORMATS[] = {
  NONE, NONE, NULL,
  NONE, NONE, NULL,
  NONE, OPEN_DIR,  NULL,
  NONE, OPEN_EXEC, NULL,
  NONE, OPEN_FIFO, NULL,
  NONE, OPEN_SOCK, NULL,
  NONE, OPEN_LINK, NULL,
  NONE, NONE, NULL,
  NONE, OPEN_DIR,  NULL,
  NONE, OPEN_EXEC, NULL,
  NONE, OPEN_FIFO, NULL,
  NONE, OPEN_SOCK, NULL,
  NONE, OPEN_LINK, NULL,
};


int NameLenAscii(const char *name) {
  return strlen(name);
}

int NameLenOctal(const char *name) {
  int len;

  for (len = 0;  *name;  ++name)
    if (' ' <= *name && *name <= '~')
      ++len;
    else
      len += 4;
  return len;
}

int NameLenQuestion(const char *name) {
  return strlen(name);
}


int PrintNameAscii(const char *name) {
  fputs(name, stdout);
  return 0;
}

int PrintNameOctal(const char *name) {
  for (;  *name;  ++name)
    if (' ' <= *name && *name <= '~')
      putc(*name, stdout);
    else
      printf("\\%03o", *name);
  return 0;
}

int PrintNameQuestion(const char *name) {
  for (;  *name;  ++name)
    if (' ' <= *name && *name <= '~')
      putc(*name, stdout);
    else
      putc('?', stdout);
  return 0;
}


/* Return number of decimal digits in num. */
static int NumberLen(int num) {
  float fnum = (float)num;
  int len = 0;

  while (fnum > 0.0F) {
    ++len;
    fnum /= 10.0F;
  }
  return len;
}


static long Blocks(const struct clsent *entries, int num) {
  long blocks = 0;
  for (;  num;  --num, ++entries)
    blocks += entries->blocks;
  return blocks;
}


static int FormatNoTypeLink(int len, const struct clsent *e) {
  fputs(FORMATS[e->type], stdout);
  PrintName(e->name);
  fputs(FORMATS[e->type + 1], stdout);
  return len + e->len;
}

static int FormatType(int len, const struct clsent *e) {
  clstype offs;

  if (e->type != SymLink)  offs = e->type;
  else                     offs = SymLink + e->target_type;
  fputs(TYPE_FORMATS[offs], stdout);
  PrintName(e->name);
  fputs(TYPE_FORMATS[offs + 1], stdout);
  return len + e->len + TYPE_LENGTHS[offs];
}

static int FormatLink(int len, const struct clsent *e) {
  clstype offs;

  if (e->type != SymLink)  offs = e->type;
  else                     offs = e->target_type;
  fputs(FORMATS[offs], stdout);
  PrintName(e->name);
  fputs(FORMATS[offs + 1], stdout);
  return len + e->len;
}

static int FormatTypeLink(int len, const struct clsent *e) {
  clstype offs;

  if (e->type != SymLink)  offs = e->type;
  else                     offs = e->target_type;
  fputs(TYPE_FORMATS[offs], stdout);
  PrintName(e->name);
  fputs(TYPE_FORMATS[offs + 1], stdout);
  return len + e->len + TYPE_LENGTHS[offs];
}


/* Stream format */
static const int RIGHT_MARGIN = 6;
static int (* FormatShort)(int, const struct clsent *);
static int (* StreamH1)(const struct clsent *);
static int (* StreamH2)(const struct clsent *, int, int);

static int Comma(int len, int total, int columns) {
  if (total + len >= columns) {
    fputs(",\n", stdout);
    total = len;
  }
  else {
    fputs(", ", stdout);
    total += len + 2;
  }
  return total;
}

static int StreamNoOp(const struct clsent *entry) {
  return 0;
}

static int Stream1(const struct clsent *entry) {
  return FormatShort(0, entry);
}

static int Stream2(const struct clsent *entry, int total, int columns) {
  int len = entry->len + (f_type ? TYPE_LENGTHS[entry->type] : 0);
  total = Comma(len, total, columns);
  FormatShort(0, entry);
  return total;
}

static int StreamInode1(const struct clsent *entry) {
  int len = printf("%ld ", entry->ino);
  return FormatShort(len, entry);
}

static int StreamInode2(const struct clsent *entry, int total, int columns) {
  int len = NumberLen(entry->ino) + 1 + entry->len +
            (f_type ? TYPE_LENGTHS[entry->type] : 0);
  total = Comma(len, total, columns);
  printf("%ld ", entry->ino);
  FormatShort(0, entry);
  return total;
}

static int StreamBlocks2(const struct clsent *entry, int total, int columns) {
  int len = NumberLen(entry->blocks) + 1 + entry->len +
            (f_type ? TYPE_LENGTHS[entry->type] : 0);
  total = Comma(len, total, columns);
  printf("%ld ", entry->blocks);
  FormatShort(0, entry);
  return total;
}

static int StreamInodeBlocks2(const struct clsent *entry, int total,
                              int columns) {
  int len = NumberLen(entry->ino) + 1 + NumberLen(entry->blocks) + 1 +
            entry->len + (f_type ? TYPE_LENGTHS[entry->type] : 0);
  total = Comma(len, total, columns);
  printf("%ld %ld ", entry->ino, entry->blocks);
  FormatShort(0, entry);
  return total;
}

void PrintStream(const struct clsent *entries, int num, int max_len,
                 int columns) {
  int i, len;

  if (num) {
    columns -= RIGHT_MARGIN;
    if (f_total) {
      len = printf("total %ld", Blocks(entries, num));
      i = 0;
    }
    else {
      len = StreamH1(&entries[0]);
      i = 1;
    }
    for (;  i < num;  i++)
      len = StreamH2(&entries[i], len, columns);
    puts("");
  }
}


/* Single column format */
static void (* SingleColH)(const struct clsent *);

void SingleCol(const struct clsent *entry) {
  FormatShort(0, entry);
  fputs("\n", stdout);
}

void SingleColInode(const struct clsent *entry) {
  printf("%6ld ", entry->ino);
  FormatShort(0, entry);
  fputs("\n", stdout);
}

void SingleColBlocks(const struct clsent *entry) {
  printf("%4ld ", entry->blocks);
  FormatShort(0, entry);
  fputs("\n", stdout);
}

void SingleColInodeBlocks(const struct clsent *entry) {
  printf("%6ld %4ld ", entry->ino, entry->blocks);
  FormatShort(0, entry);
  fputs("\n", stdout);
}

void PrintSingleCol(const struct clsent *entries, int num,
                    int max_len, int columns) {
  int i;

  if (f_total)
    printf("total %ld\n", Blocks(entries, num));
  for (i = 0;  i < num;  i++)
    SingleColH(&entries[i]);
}


/* Multi-column, sorted down the columns format */
static const int MIN_COL_LEN      =  8;
static const int INODE_LEN        =  7;
static const int BLOCKS_LEN       =  5;
static const int INODE_BLOCKS_LEN = 12;
static int COL_LEN_ADD            =  2;
static char SPACE[MAX_PATH + 1];
static int (* ColumnH)(const struct clsent *);

static int Column(const struct clsent *entry) {
  return FormatShort(0, entry);
}

static int ColumnInode(const struct clsent *entry) {
  int len = printf("%6ld ", entry->ino);
  return FormatShort(len, entry);
}

static int ColumnBlocks(const struct clsent *entry) {
  int len = printf("%4ld ", entry->blocks);
  return FormatShort(len, entry);
}

static int ColumnInodeBlocks(const struct clsent *entry) {
  int len = printf("%6ld %4ld ", entry->ino, entry->blocks);
  return FormatShort(len, entry);
}

void PrintColDown(const struct clsent *entries, int num, int max_len,
                  int columns) {
  int cols, rows, len, i, j, k, start_i, num_space;

  if ((max_len += COL_LEN_ADD) < MIN_COL_LEN)
    max_len = MIN_COL_LEN;
  if (f_blocks && f_total) {
    /* Output total number of blocks. */
    len = printf("total %ld", Blocks(entries, num));
    if (len > max_len)
      max_len = len + 1;
    start_i = 1;
    num_space = max_len - len;
  }
  else
    start_i = num_space = 0;

  cols = max_len < columns ? columns / max_len : 1;
  rows = (num + start_i) / cols + ((num + start_i) % cols ? 1 : 0);
  for (j = 0;  j < rows;  ++j) {
    for (i = j == 0 ? start_i : 0;  i < cols;  ++i) {
      if ((k = i * rows + j - start_i) < num) {
        if (num_space)
          printf("%.*s", num_space, SPACE);
        num_space = max_len - ColumnH(&entries[k]);
      }
      else
        break;
    }
    puts("");
    num_space = 0;
  }

#ifdef WORKING
  if (!f_blocks) {
    cols = max_len < columns ? columns / max_len : 1;
    rows = num / cols + (num % cols ? 1 : 0);
    for (j = 0;  j < rows;  j++) {
      num_space = 0;
      for (i = 0;  i < cols;  i++) {
        if ((k = i * rows + j) < num) {
          if (num_space)
            printf("%.*s", num_space, SPACE);
          num_space = max_len - ColumnH(&entries[k]);
        }
        else
          break;
      }
      puts("");
    }
  }
  else {
    /* If f_blocks then we output total number of blocks. */
    len = printf("total %ld", Blocks(entries, num));
    if (len > max_len)
      max_len = len + 1;

    cols = max_len < columns ? columns / max_len : 1;
    rows = (num + 1) / cols + ((num + 1) % cols ? 1 : 0);
    num_space = max_len - len;
    for (i = 1;  i < cols;  i++) {
      if ((k = i * rows - 1) < num) {
        if (num_space)
          printf("%.*s", num_space, SPACE);
        num_space = max_len - ColumnH(&entries[k]);
      }
      else
        break;
    }
    puts("");
    for (j = 1;  j < rows;  j++) {
      num_space = 0;
      for (i = 0;  i < cols;  i++) {
        if ((k = i * rows + j - 1) < num) {
          if (num_space)
            printf("%.*s", num_space, SPACE);
          num_space = max_len - ColumnH(&entries[k]);
        }
        else
          break;
      }
      puts("");
    }
  }
#endif
}


/* Multi-column, sorted across the columns format */
void PrintColAcross(const struct clsent *entries, int num, int max_len,
                    int columns) {
  int cols, rows, len, i, k, start_i, num_space;

  if ((max_len += COL_LEN_ADD) < MIN_COL_LEN)
    max_len = MIN_COL_LEN;
  if (f_blocks && f_total) {
    /* Output total number of blocks. */
    len = printf("total %ld", Blocks(entries, num));
    if (len > max_len)
      max_len = len + 1;
    start_i = 1;
    num_space = max_len - len;
  }
  else
    start_i = num_space = 0;
  k = 0;

  /* Output full rows. */
  cols = max_len < columns ? columns / max_len : 1;
  for (rows = (num + start_i) / cols;  rows;  --rows) {
    for (i = start_i;  i < cols;  ++i, ++k) {
      if (num_space)
        printf("%.*s", num_space, SPACE);
      num_space = max_len - ColumnH(&entries[k]);
    }
    puts("");
    start_i = num_space = 0;
  }

  /* Output remaining files on last line. */
  if (k < num) {
    do {
      if (num_space)
        printf("%.*s", num_space, SPACE);
      num_space = max_len - ColumnH(&entries[k]);
    } while (++k < num);
    puts("");
  }
}


/* Long format */
static time_t SIX_MONTHS_AGO;
static void (* LongInodeH)(const struct clsent *entry);
static void (* LongBlocksH)(const struct clsent *entry);
static void (* LongUserH)(const struct clsent *entry);
static void (* LongGroupH)(const struct clsent *entry);
static void (* LongTimeH)(const struct clsent *entry);
static void (* LongNameH)(const struct clsent *entry);

static char *PermsString(char *s, mode_t m) {
  if (S_ISLNK(m))       s[0] = 'l';
  else if (S_ISREG(m))  s[0] = '-';
  else if (S_ISDIR(m))  s[0] = 'd';
  else if (S_ISCHR(m))  s[0] = 'c';
  else if (S_ISBLK(m))  s[0] = 'b';
  else if (S_ISFIFO(m)) s[0] = 'p';
  else if (S_ISSOCK(m)) s[0] = 's';
#ifdef _HPUX_SOURCE
  else if (S_ISNWK(m))  s[0] = 'n';
#endif
  else                  s[0] = '-';

  s[1] = m & S_IRUSR ? 'r' : '-';
  s[2] = m & S_IWUSR ? 'w' : '-';
  if (m & S_IXUSR) s[3] = m & S_ISUID ? 's' : 'x';
  else             s[3] = m & S_ISUID ? 'S' : '-';

  s[4] = m & S_IRGRP ? 'r' : '-';
  s[5] = m & S_IWGRP ? 'w' : '-';
  if (m & S_IXGRP) s[6] = m & S_ISGID ? 's' : 'x';
  else             s[6] = m & S_ISGID ? 'S' : '-';

  s[7] = m & S_IROTH ? 'r' : '-';
  s[8] = m & S_IWOTH ? 'w' : '-';
  if (m & S_IXOTH) s[9] = m & 0001000 ? 't' : 'x';
  else             s[9] = m & 0001000 ? 'T' : '-';

  s[10] = '\0';
  return s;
}

static void LongNoOp(const struct clsent *entry) {
}

static void LongInode(const struct clsent *entry) {
  printf("%6ld ", entry->ino);
}

static void LongBlocks(const struct clsent *entry) {
  printf("%4ld ", entry->blocks);
}

static void LongUser(const struct clsent *entry) {
  if (entry->user) printf("%-8s ", entry->user);
  else             printf("%-8d ", entry->uid);
}

static void LongGroup(const struct clsent *entry) {
  if (entry->group) printf("%-8s ", entry->group);
  else              printf("%-8d ", entry->gid);
}

static void LongNumUser(const struct clsent *entry) {
  printf("%-8d ", entry->uid);
}

static void LongNumGroup(const struct clsent *entry) {
  printf("%-8d ", entry->gid);
}

#define PRINT_TIME(field) \
  time_t t = entry->field; \
  const char *a = ctime(&t); \
  if (t > SIX_MONTHS_AGO) \
    printf("%.6s %.5s ",  a + 4, a + 11); \
  else \
    printf("%.6s  %.4s ", a + 4, a + 20) \

static void LongModTime(const struct clsent *entry) {
  PRINT_TIME(mtime);
}

static void LongAccessTime(const struct clsent *entry) {
  PRINT_TIME(atime);
}

static void LongStatusTime(const struct clsent *entry) {
  PRINT_TIME(ctime);
}

static void FormatLong(clstype type, const char *name) {
  fputs(LONG_FORMATS[type], stdout);
  PrintName(name);
  fputs(LONG_FORMATS[type + 1], stdout);
}

static void LongName(const struct clsent *entry) {
  FormatLong(entry->type, entry->name);
  if (entry->type == SymLink) {
    fputs(" -> ", stdout);
    FormatLong(entry->target_type, entry->target_name);
  }
}

static void LongLinkName(const struct clsent *entry) {
  if (entry->type == SymLink)
    FormatLong(entry->target_type, entry->name);
  else
    FormatLong(entry->type, entry->name);
}

void PrintLong(const struct clsent *entries, int num, int max_len,
               int columns) {
  const struct clsent *e;
  long blocks;
  int i;
  char perms[11];

  if (f_total)
    printf("total %ld\n", Blocks(entries, num));
  for (i = num;  i;  i--, entries++) {
    LongInodeH(entries);
    LongBlocksH(entries);
    printf("%s %3d ", PermsString(perms, entries->mode), entries->nlink);
    LongUserH(entries);
    LongGroupH(entries);
    if (S_ISCHR(entries->mode) || S_ISBLK(entries->mode))
      printf("%3ld 0x%.6lx ", (entries->rdev & 0xff000000UL) >> 24,
             entries->rdev & 0xffffffUL);
    else
      printf("%7ld ", entries->size);
    LongTimeH(entries);
    LongNameH(entries);
    puts("");
  }
}


void PrintInit() {
  SIX_MONTHS_AGO = time(NULL) - (365 / 2) * 24 * 60 * 60;
  memset(SPACE, ' ', sizeof(SPACE));
  SPACE[sizeof(SPACE) - 1] = '\0';

  /* Select name length and printing functions. */
  switch (f_nongraphic) {
    case Ascii:    NameLen = NameLenAscii;
                   PrintName = PrintNameAscii;
                   break;
    case Octal:    NameLen = NameLenOctal;
                   PrintName = PrintNameOctal;
                   break;
    case Question: NameLen = NameLenQuestion;
                   PrintName = PrintNameQuestion;
                   break;
  }

  /* Select name formating functions. */
  if (f_printform != Long) {
    if (f_type && f_link)
      FormatShort = FormatTypeLink;
    else if (f_type)
      FormatShort = FormatType;
    else if (f_link)
      FormatShort = FormatLink;
    else
      FormatShort = FormatNoTypeLink;
  }

  switch (f_term) {
    case Dtterm:
      FORMATS      = DT_FORMATS;
      TYPE_FORMATS = DT_TYPE_FORMATS;
      LONG_FORMATS = f_type ? DT_TYPE_FORMATS : DT_FORMATS;
      break;
    case Hpterm:
      FORMATS      = HP_FORMATS;
      TYPE_FORMATS = HP_TYPE_FORMATS;
      LONG_FORMATS = f_type ? HP_TYPE_FORMATS : HP_FORMATS;
      break;
    default:
      FORMATS      = OPEN_FORMATS;
      TYPE_FORMATS = OPEN_TYPE_FORMATS;
      LONG_FORMATS = f_type ? OPEN_TYPE_FORMATS : OPEN_FORMATS;
      break;
  }

  /* Select a print function. */
  switch (f_printform) {
    case ColumnDown:
      if (f_inode && f_blocks) {
        COL_LEN_ADD += INODE_BLOCKS_LEN;
        ColumnH = ColumnInodeBlocks;
      }
      else if (f_inode) {
        COL_LEN_ADD += INODE_LEN;
        ColumnH = ColumnInode;
      }
      else if (f_blocks) {
        COL_LEN_ADD += BLOCKS_LEN;
        ColumnH = ColumnBlocks;
      }
      else
        ColumnH = Column;
      PrintFunc = PrintColDown;
      break;
    case Long:
      LongInodeH  = f_inode  ? LongInode  : LongNoOp;
      LongBlocksH = f_blocks ? LongBlocks : LongNoOp;
      if (f_numbers) {
        LongUserH  = f_noowner ? LongNoOp : LongNumUser;
        LongGroupH = f_nogroup ? LongNoOp : LongNumGroup;
      }
      else {
        LongUserH  = f_noowner ? LongNoOp : LongUser;
        LongGroupH = f_nogroup ? LongNoOp : LongGroup;
      }
      switch (f_usetime) {
        case Modification: LongTimeH = LongModTime;     break;
        case Status:       LongTimeH = LongStatusTime;  break;
        case Access:       LongTimeH = LongAccessTime;  break;
        default:           LongTimeH = LongNoOp;       break;
      }
      LongNameH = f_link ? LongLinkName : LongName;
      PrintFunc = PrintLong;
      break;
    case SingleColumn:
      if (f_inode && f_blocks)
        SingleColH = SingleColInodeBlocks;
      else if (f_inode)
        SingleColH = SingleColInode;
      else if (f_blocks)
        SingleColH = SingleColBlocks;
      else
        SingleColH = SingleCol;
      PrintFunc = PrintSingleCol;
      break;
    case ColumnAcross:
      if (f_inode && f_blocks) {
        COL_LEN_ADD += INODE_BLOCKS_LEN;
        ColumnH = ColumnInodeBlocks;
      }
      else if (f_inode) {
        COL_LEN_ADD += INODE_LEN;
        ColumnH = ColumnInode;
      }
      else if (f_blocks) {
        COL_LEN_ADD += BLOCKS_LEN;
        ColumnH = ColumnBlocks;
      }
      else
        ColumnH = Column;
      PrintFunc = PrintColAcross;
      break;
    case Stream:
      if (f_inode && f_blocks) {
        StreamH1 = StreamNoOp;
        StreamH2 = StreamInodeBlocks2;
      }
      else if (f_inode) {
        StreamH1 = StreamInode1;
        StreamH2 = StreamInode2;
      }
      else if (f_blocks) {
        StreamH1 = StreamNoOp;
        StreamH2 = StreamBlocks2;
      }
      else {
        StreamH1 = Stream1;
        StreamH2 = Stream2;
      }
      PrintFunc = PrintStream;
      break;
  }
}


void PrintDir(const char *name) {
  int len = strlen(name);

  if (name[len - 1] != '.' || (len > 1 && name[len - 2] != '/')) {
    FormatLong(Directory, name);
    puts(":");
  }
}
