/* Maintain a cache of user and group names.
   Copyright (C) 1998 Rex Barzee

   This file is part of cls which lists contents of directories in color.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "cls.h"
#include <grp.h>
#include <pwd.h>
#include <stdlib.h>
#include <string.h>

#define PW_CACHE 90
#define GR_CACHE 50

struct pwent {
  uid_t pw_uid;
  const char *pw_name;
};

struct grent {
  gid_t gr_gid;
  const char *gr_name;
};

static struct pwent passwd[PW_CACHE];
static struct grent group[GR_CACHE];
static int num_passwd, num_group;


static int UidCmp(const void *pw1, const void *pw2) {
  return ((struct pwent *)pw1)->pw_uid - ((struct pwent *)pw2)->pw_uid;
}


static const char *AddUser(uid_t uid) {
  struct passwd *p;
  const char *name;

  if (p = getpwuid(uid)) {
    if (num_passwd < PW_CACHE) {
      passwd[num_passwd].pw_uid = p->pw_uid;
      passwd[num_passwd].pw_name = name = strdup(p->pw_name);
      num_passwd++;
      qsort(passwd, num_passwd, sizeof(*passwd), UidCmp);
      return name;
    }
    return p->pw_name;
  }
  return NULL;
}


const char *GetUser(uid_t uid) {
  if (f_printform == Long && !f_noowner) {
    struct pwent key, *p;

    key.pw_uid = uid;
    p = (struct pwent *)bsearch(&key, passwd, num_passwd,
                                sizeof(*passwd), UidCmp);
    return p ? p->pw_name : AddUser(uid);
  }
  return NULL;
}


#if 0
/* No reason to free this memory because it would only be freed when the
   the program is exiting. */
void UserDestroy() {
  struct pwent *p = passwd;
  int i;

  for (i = num_passwd;  i;  --i, ++p)
    free((void *)p->pw_name);
}
#endif


static int GidCmp(const void *gr1, const void *gr2) {
  return ((struct grent *)gr1)->gr_gid - ((struct grent *)gr2)->gr_gid;
}


static const char *AddGroup(gid_t gid) {
  struct group *g;
  const char *name;

  if (g = getgrgid(gid)) {
    if (num_group < GR_CACHE) {
      group[num_group].gr_gid = g->gr_gid;
      group[num_group].gr_name = name = strdup(g->gr_name);
      num_group++;
      qsort(group, num_group, sizeof(*group), GidCmp);
      return name;
    }
    return g->gr_name;
  }
  return NULL;
}


const char *GetGroup(gid_t gid) {
  if (f_printform == Long && !f_nogroup) {
    struct grent key, *g;

    key.gr_gid = gid;
    g = (struct grent *)bsearch(&key, group, num_group,
                                sizeof(*group), GidCmp);
    return g ? g->gr_name : AddGroup(gid);
  }
  return NULL;
}


#if 0
/* No reason to free this memory because it would only be freed when the
   the program is exiting. */
void GroupDestroy() {
  struct grent *g = group;
  int i;

  for (i = num_group;  i;  --i, ++g)
    free((void *)g->gr_name);
}
#endif
